!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!
!!! Copyright (C) 2007 Jason D. Nielsen
!!!
!!! This program is free software! you can redistribute it and/or modify
!!! it under the terms of the GNU General Public License as published by
!!! the Free Software Foundation! either version 2 of the License, or
!!! (at your option) any later version.
!!!
!!! This program is distributed in the hope that it will be useful,
!!! but WITHOUT ANY WARRANTY! without even the implied warranty of
!!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!! GNU General Public License for more details.
!!!
!!! You should have received a copy of the GNU General Public License
!!! along with this program! if not, write to the Free Software
!!! Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA
!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

include 'depends.f90'

module stp_dm_NHPP_mod
  use mytype
  implicit none
  private
  public:: stpdmNHPP
  integer, parameter:: nmaxit=1000
  real(dp), allocatable, save:: nij(:,:), mui(:), ni(:),  pdiff(:), probs(:), dlike(:), rscl(:), gmln(:)
  real(dp), allocatable, save, target:: coef_t(:,:), pr_wt_t(:,:), ebeta_t(:,:)
  real(dp), pointer:: beta(:) => null(), phi(:) => null(), tau => null(), pr_wt(:) => null(), &
       & ebeta(:) => null()
  real(dp), save:: loglike
  logical, save, allocatable:: nijNA(:,:)
  integer, save, allocatable:: phld(:,:)
  integer, save:: nn, np, ncf, ngr, err

contains
  
  subroutine init(NData,Xm,Pan,pwt,nnn,pp,cc,gg)
    use mytype
    use dm_util, only: gammln, set_random_seed
    implicit none 
    integer, intent(in):: nnn, pp, cc, gg, Xm(cc)
    real(dp), intent(in):: NData(pp,nnn), Pan(pp)
    real(dp), intent(inout):: pwt(nnn,gg)
    integer:: i, Xvc(cc)
    nn=nnn
    np=pp
    ncf=cc
    ngr=gg
    err=0
    allocate(nij(nn,np),mui(nn),ni(nn),pdiff(np),ebeta_t(nn,ngr),phld(2,ncf), &
         & coef_t(ncf+np,ngr), pr_wt_t(nn,ngr), probs(ngr), rscl(nn), &
         & dlike(nn), gmln(nn), nijNA(nn,np))
    nij=transpose(NData)
    Xvc=Xm
    pr_wt_t=pwt
    pdiff=Pan-(/0.0_dp,Pan(1:(np-1))/)
    do i=2,ncf
       Xvc(i)=Xvc(i)+Xvc(i-1)
    end do
    phld(2,:)=Xvc
    phld(1,:)=(/0,phld(2,1:(ncf-1))/)+1.0_dp
    probs=sum(pr_wt_t,1)/real(nn,dp)
    nijNA=(nij > -0.01_dp)
    where (.not. nijNA) 
       nij=0.0_dp
    end where
    ni=sum(nij,2)
    do i=1,nn
       gmln(i)=sum(gammln(nij(i,:)+1.0_dp))
    end do
    call set_random_seed()
  end subroutine init
   
  subroutine close
    implicit none 
    deallocate(nij,mui,ni,pdiff,ebeta_t,phld,coef_t,pr_wt_t,probs,rscl,dlike,gmln, &
         & nijNA)
    nullify(beta,phi,tau,pr_wt,ebeta)
  end subroutine close

  subroutine init_prb
    use mytype
    implicit none
    real(dp):: tprb(nn), t_pr(nn,ngr), bfun, bfun_o
    integer:: i, j, k
    bfun=-100000.0_dp
    bfun_o=bfun
    do i=1,nmaxit
       if (bfun > bfun_o) then
          bfun_o=bfun
       end if
       do j=1,50*ngr
          call random_number(pr_wt_t)
          tprb=sum(pr_wt_t,2)
          do k=1,ngr
             pr_wt_t(:,k)=pr_wt_t(:,k)/tprb
          end do
          probs=sum(pr_wt_t,1)/real(nn,dp)
          call es_update
          if (err /= 0) then
             err=0             
             cycle             
          end if
          if (loglike > bfun) then
             bfun=loglike
             t_pr=pr_wt_t
          end if
       end do
       if (abs(bfun-bfun_o)< 0.00001_dp ) then
          pr_wt_t=t_pr
          return
       end if
    end do
  end subroutine init_prb

  function tau_est(x)
    use mytype
    implicit none
    real(dp), intent(in):: x
    real(dp):: tau_est
    real(dp):: nivar(nn), tmui(nn)
    tmui=1.0_dp+x*mui
    nivar=mui*tmui
    tau_est=sum(pr_wt*(((ni-mui)**2)/nivar + log(nivar)))
  end function tau_est

  subroutine s_step
    use mytype
    use dm_util, only: fmin
    implicit none
    real(dp):: fval, upb, eb_tmp(ncf), pnij(nn,np)
    integer:: i, j, k
    do j=1,ngr
       pr_wt => pr_wt_t(:,j)
       ebeta => ebeta_t(:,j)
       beta => coef_t(1:(ncf-1),j)
       phi => coef_t(ncf:(ncf+np-1),j)
       tau => coef_t(np+ncf,j)
       tau=0.0_dp
       upb=1.0_dp
       do i=1,np
          pnij(:,i)=nij(:,i)*pr_wt
       end do
       do i=1,ncf
          eb_tmp(i)=sum(pnij(phld(1,i):phld(2,i),:),mask=nijNA(phld(1,i):phld(2,i),:))
          fval=sum(pr_wt(phld(1,i):phld(2,i)))
          if (fval < 1) then
             err=-99
             return
          end if          
          eb_tmp(i)=eb_tmp(i)/fval
       end do
       eb_tmp=eb_tmp/eb_tmp(1)
       do i=1,ncf
          ebeta(phld(1,i):phld(2,i))=eb_tmp(i)
       end do
       phi=sum(pnij,1,nijNA)/sum(pr_wt*ebeta)
       where (phi < 1.0e-20_dp) 
          phi=1.0e-20_dp
       end where
       do k=1,nn
          mui(k)=sum(phi,mask=nijNA(k,:))*ebeta(k)
       end do
       do i=1,20
          call fmin(0.0001_dp,upb,tau,fval,tau_est,1.0e-6_dp)
          if (tau < 0.98*upb) then
             exit
          end if
          upb=upb+1.0_dp
       end do
       beta=log(eb_tmp(2:ncf))
       phi=log(phi/pdiff)
    end do
  end subroutine s_step

  subroutine e_step
    use mytype
    use dm_util, only: poisson
    implicit none
    integer:: i, j, k
    real(dp):: dd(np), vv(nn), dplike(nn,ngr), tphi(np)
    do i=1,ngr
       phi => coef_t(ncf:(ncf+np-1),i)
       tau => coef_t(ncf+np,i)
       ebeta => ebeta_t(:,i)
       tphi=exp(phi)
       tphi=pdiff*tphi
       where (tphi < 1.0e-20_dp) 
          tphi=1.0e-20_dp
       end where
       do j=1,nn
          mui(j)=sum(phi,mask=nijNA(j,:))*ebeta(j)
       end do
       mui=sum(tphi)*ebeta
       vv=(1.0_dp+tau*ni)/(1.0_dp+tau*mui)
       do j=1,nn     
          do k=1,np
             if (nijNA(j,k)) then
                dd(k)=poisson(nij(j,k),vv(j)*tphi(k)*ebeta(j))
             else
                dd(k)=0.0_dp
             end if
          end do
          dplike(j,i)=log(probs(i))+sum(dd)
       end do       
    end do
    rscl=maxval(dplike,2)
    do i=1,ngr
       dplike(:,i)=exp(dplike(:,i)-rscl)
    end do
    dlike=sum(dplike,2)
    do i=1,ngr
       pr_wt_t(:,i)=dplike(:,i)/dlike
    end do
    probs=sum(pr_wt_t,1)/real(nn,dp)
  end subroutine e_step

  subroutine es_update
    use mytype
    implicit none
    real(dp):: loglikeo
    integer:: i
    call s_step
    if (err /= 0) return
    call e_step
    loglike=sum(log(dlike)+rscl-gmln)
    loglikeo=loglike
    do i=1,nmaxit
       call s_step
       if (err /= 0) return
       call e_step
       loglike=sum(log(dlike)+rscl-gmln)
       if (abs(loglike-loglikeo)<0.00001_dp) then
          exit
       end if
       loglikeo=loglike
    end do
  end subroutine es_update

  subroutine stpdmNHPP(NData,Xm,Pan,pwt,nnn,pp,cc,gg,ierr)
    use mytype
    implicit none
    integer, intent(in):: nnn, pp, cc, gg, Xm(cc)
    real(dp), intent(in):: NData(pp,nnn), Pan(pp)
    real(dp), intent(inout):: pwt(nnn,gg)
    integer, intent(out):: ierr
    call init(NData,Xm,Pan,pwt,nnn,pp,cc,gg)
    call init_prb
    call es_update
    if (err /= 0) then
       ierr=-99
       call close
       return
    end if
    pwt=pr_wt_t
    ierr=0
    call close
  end subroutine stpdmNHPP

end module stp_dm_NHPP_mod

module dm_NHPP_mod
  use mytype
  implicit none
  private
  public:: dmNHPP
  real(dp), save, allocatable:: Bmat(:,:), tpanel(:), time_grid(:), &
       & w_grid(:), ni(:), nij(:,:), mui(:), pDVD(:,:), DVD(:,:), Scr(:,:), &
       & PenM(:,:), probs(:), tBBB(:,:), tdiff(:), gmln(:), ddlike(:,:), &
       & dplike(:,:), dlike(:), rscl(:)
  real(dp), save, allocatable, target:: mu_t(:,:), Dmu_t(:,:), coef_t(:,:), &
       & pr_wt_t(:,:), Scr_t(:,:)
  real(dp), pointer:: muij(:,:) => null(), Dmu(:,:) => null(), psi(:) => null(), &
       & spar(:) => null(), pr_wt(:) => null(), tau => null(), Scr_p(:,:) => null()
  logical, save, allocatable:: nijNA(:,:)
  real(dp), save:: w_qdr(4), a_qdr(4), loglike, eff_df
  integer, save, allocatable:: Itrt(:), Xvc(:), sqpan(:)
  integer, save:: nind, npan, nkn, ntrt, nXr, nBc, ntBc, ngrp, nBgrp, nBngr, err, &
       & fini, npgrp, mngrp
  integer, parameter:: maxit=150, nmaxit=200, pmaxit=2
  character(len=*), parameter:: form1="(A20,1X,F12.3)"

contains

!@init

subroutine initialize(NData,Xm,Panel,pwt,npn,nsub,nknt,ntr,ngr,fin)
  use mytype
  use dm_util, only: gammln, bbase, diffd
  use matrix, only: operator(.kp.), operator(.mm.), operator(.tmm.), finverse
  implicit none
  integer, intent(in):: npn, nsub, nknt, ntr, ngr, fin
  real(dp), intent(in):: NData(npn,nsub), Panel(npn), pwt(nsub,ngr)
  integer, intent(in):: Xm(ntr)
  integer:: i, mnBngr, mnpgrp
  real(dp):: sqr_thirty, expnd(nsub), d_tpanel(npn), m_tpanel(npn), smpr(ntr), &
       & DD(nknt+3,nknt+4), BB(npn,nknt+4),tBB(nsub*npn,ntr*(nknt+4)), &
       & BBB(ntr*(nknt+4),ntr*(nknt+4))
  err=0
! Maximum number of clusters/subgroups(populations)
  mngrp=15
  if (ngr > 15) then
     write(*,'(A31,I2,A1)') 'Number of groups cannot exceed ',mngrp,'!'
     err=999
     return
  end if
  fini=fin
  nind=nsub
  npan=npn
  ngrp=ngr
  ntrt=ntr
  nXr=nind*npan
  npgrp=ngrp*npan
  mnpgrp=mngrp*npan
  nkn=nknt
  nBc=nkn+4
  ntBc=ntrt*nBc
  nBgrp=ngrp*ntBc+ngrp
  nBngr=nBgrp+ngrp-1
  mnBngr=mngrp*ntBc+2*mngrp-1
! Set up abscissas and weights for Legendre-Gauss Quadrature (4 point - exact
! for polynomials of degree 7 or less
!-------------------------------------------------------------------------  
  sqr_thirty=sqrt(30.0_dp)
  w_qdr(1)=(18.0_dp-sqr_thirty)/36.0_dp
  w_qdr(2)=(18.0_dp+sqr_thirty)/36.0_dp
  w_qdr(3)=w_qdr(2)
  w_qdr(4)=w_qdr(1)
  a_qdr(4)=sqrt(525.0_dp +(70.0_dp*sqr_thirty))/35.0_dp 
  a_qdr(3)=sqrt(525.0_dp -(70.0_dp*sqr_thirty))/35.0_dp
  a_qdr(2)=-a_qdr(3)
  a_qdr(1)=-a_qdr(4)
!-------------------------------------------------------------------------
  allocate(Xvc(ntrt), Itrt(nind), Bmat(4*npan,nBc), time_grid(4*npan), w_grid(4*npan), &
       & tpanel(npan), coef_t(ntBc+ntrt+1,mngrp), nij(npan,nind), ni(nind), mui(nind), &
       & nijNA(npan,nind), Dmu_t(mnpgrp,ntBc), DVD(ntBc,ntBc), pDVD(ntBc,ntBc), &
       & Scr(ntBc,nind), PenM(nBc,nBc), pr_wt_t(nind,mngrp), probs(mngrp), &
       & dlike(nind), ddlike(nind,mngrp), dplike(nind,mngrp), mu_t(mnpgrp,nind), &
       & tBBB(ntBc,nXr), tdiff(nXr), gmln(nind), rscl(nind), Scr_t(mnBngr,nind), &
       & sqpan(npan))
!------------------------------------------------
  nij=NData
  tpanel=Panel/Panel(npan) 
  nijNA=.FALSE.
  where(nij< 0)
     nijNA=.TRUE.
     nij=0.0_dp
  end where
  ni=sum(nij,1)
  sqpan=(/(i,i=1,npan)/)
  Xvc=Xm
  Itrt(1:Xvc(1))=1
  if (ntrt > 1) then
     do i=2,ntrt
        Itrt((Xvc(i-1)+1):sum(Xvc(1:i)))=i
     end do
  end if
  d_tpanel=(tpanel-(/0.0_dp,tpanel(1:(npan-1))/))/2.0_dp
  m_tpanel=(tpanel+(/0.0_dp,tpanel(1:(npan-1))/))/2.0_dp
  time_grid(1:4)=a_qdr*d_tpanel(1)+m_tpanel(1)
  w_grid(1:4)=w_qdr*d_tpanel(1)*Panel(npan)
  do i=2,npan
     time_grid(((i-1)*4+1):(i*4))=a_qdr*d_tpanel(i) + m_tpanel(i)
     w_grid(((i-1)*4+1):(i*4))=w_qdr*d_tpanel(i)*Panel(npan)
  end do
  Bmat=bbase(time_grid,0.0_dp,1.0_dp,nkn,3)
  DD=diffd(nBc,1)
  PenM=DD.tmm.DD
  d_tpanel=Panel(npan)*d_tpanel*2.0_dp
!------------------------------------------------
  pr_wt_t(:,1:ngrp)=pwt
  probs(1:ngrp)=sum(pr_wt_t(1:nind,1:ngrp),1)/real(nind,dp)
!------------------------------------------------
  tdiff=pack(spread(log(d_tpanel),2,nind),.TRUE.)
  where(pack(nijNA,.true.)) tdiff=0.0_dp
  BB=bbase(m_tpanel,0.0_dp,1.0_dp,nkn,3)
  expnd=1.0_dp
  tBB(:,(nBc+1):(nBc*ntrt))=0.0_dp
  tBB(:,1:nBc)=expnd.kp.BB
  if (ntrt > 1) then
     do i=2,ntrt
        tBB((Xvc(i-1)*npan+1):(sum(Xvc(1:i))*npan),((i-1)*nBc+1):(i*nBc))=expnd(1:Xvc(i)).kp.BB
     end do
  end if
  tBB(pack((/(i,i=1,nXr)/),pack(nijNA,.TRUE.)),:)=0.0_dp
  BBB=tBB.tmm.tBB
  smpr=real(Xvc,dp)*0.5_dp
  do i=1,ntrt
     BBB(((i-1)*nBc+1):(i*nBc),((i-1)*nBc+1):(i*nBc))=BBB(((i-1)*nBc+1):(i*nBc),&
          & ((i-1)*nBc+1):(i*nBc))+(smpr(i)*PenM)
  end do
  BBB=finverse(BBB,err)
  if (err /=0) return
  tBB=tBB.mm.BBB
  tBBB=transpose(tBB)
!------------------------------------------------
  do i=1,nind
     gmln(i)=sum(gammln(nij(:,i)+1.0_dp))
  end do
end subroutine initialize

subroutine init_param
  use mytype
  use matrix, only: operator(.mm.)
  implicit none
  real(dp):: logni1(nXr)
  logni1=pack(nij*spread(pr_wt,1,npan),.true.)
  where (logni1<0.0001_dp)
     logni1=0.001_dp
  elsewhere
     logni1=log(logni1)
  end where
  logni1=logni1-tdiff
  tau=0.0_dp
  spar=0.5_dp
  psi=tBBB.mm.logni1
end subroutine init_param

subroutine unload
  implicit none
  nullify(muij, Dmu, psi, tau, spar, pr_wt, Scr_p)
  deallocate(Xvc, Itrt, Bmat, time_grid, w_grid, tpanel, coef_t, nij, ni, mui, nijNA, &
       & Dmu_t, DVD, pDVD, Scr, PenM, pr_wt_t, probs, dlike, ddlike, dplike, mu_t, &
       & tBBB, tdiff, gmln, rscl, Scr_t, sqpan)
end subroutine unload

subroutine param_update(niter)
  use mytype
  use matrix, only: operator(.mm.), rsymsolve
  implicit none
  integer, intent(in):: niter
  real(dp):: dDVD(ntBc,ntBc), spar_old(ntrt)
  integer:: i
  call mean_update(niter)
  if (err /= 0) return
  call tau_update
  call rsymsolve(pDVD,DVD,dDVD)
  call spar_update(dDVD)
  call mean_update(niter)
  if (err /= 0) return
  call tau_update
  spar_old=log10(spar)
  do i=1,niter
     call rsymsolve(pDVD,DVD,dDVD)
     call spar_update(dDVD)
     if (maxval(abs(spar_old-log10(spar)))<1.0e-1_dp) then
        exit
     end if
     spar_old=log10(spar)
     call mean_update(niter)
     if (err /= 0) return
     call tau_update
  end do
end subroutine param_update

subroutine fparam_update(niter)
  use mytype
  use matrix, only: fsymsolve
  implicit none
  integer, intent(in):: niter
  real(dp):: dDVD(ntBc,ntBc), spar_old(ntrt)
  integer:: i
  call mean_t_update(niter)
  if (err /= 0) return
  call fsymsolve(pDVD,DVD,dDVD,err)
  if (err /= 0) return
  call spar_update(dDVD)
  spar_old=log10(spar)
  do i=1,niter
     call mean_t_update(niter)
     if (err /= 0) return
     call fsymsolve(pDVD,DVD,dDVD,err)
     if (err /= 0) return
     call spar_update(dDVD)
     if (maxval(abs(spar_old-log10(spar)))<1.0e-1_dp) then
        exit
     end if
     spar_old=log10(spar)
  end do
end subroutine fparam_update

subroutine spar_update(dDVD)
  use mytype
  use matrix, only: diag, operator(.mm.)
  implicit none
  real(dp), intent(in):: dDVD(ntBc,ntBc)
  real(dp):: tpsi(nBc), tDVD(nBc,nBc)
  integer:: i
  do i=1,ntrt
     tDVD=dDVD(((i-1)*nBc+1):(i*nBc),((i-1)*nBc+1):(i*nBc))
     tpsi=psi(((i-1)*nBc+1):(i*nBc))
     spar(i)=(sum(diag(tDVD))/sum((PenM.mm.tpsi)*tpsi))/real(Xvc(i),dp)
     if (spar(i)>1.0e4_dp) then
        spar(i)=1.0e4_dp
     end if
     if (spar(i)<1.0e-7_dp) then
        spar(i)=1.0e-7_dp
     end if
  end do
end subroutine spar_update

subroutine mean_update(niter)
  use mytype
  use dm_util, only: newton
  implicit none
  integer, intent(in):: niter
  real(dp):: coeff(ntBc), grd(ntBc), jaco(ntBc,ntBc)
  coeff=psi
  call newton(ntBc, coeff, grd, jaco, stepM, niter, err)
end subroutine mean_update

subroutine mean_t_update(niter)
  use mytype
  use dm_util, only: newton
  implicit none
  integer, intent(in):: niter
  real(dp):: coeff(ntBc+1), grd(ntBc+1), jaco(ntBc+1,ntBc+1)
  coeff(1:ntBc)=psi
  coeff(ntBc+1)=tau
  call newton(ntBc+1, coeff, grd, jaco, stepMT, niter, err)
end subroutine mean_t_update

subroutine stepM(n,x,fvec,fjac)
  use mytype
  implicit none
  integer, intent(in):: n
  real(dp), intent(in):: x(n)
  real(dp), intent(out):: fvec(n)
  real(dp), intent(out):: fjac(n,n)
  psi=x
  call update_mu_dmu
  fvec=sum(Scr,2)
  fjac=pDVD
end subroutine stepM

subroutine stepMT(n,x,fvec,fjac)
  use mytype
  implicit none
  integer, intent(in):: n
  real(dp), intent(in):: x(n)
  real(dp), intent(out):: fvec(n)
  real(dp), intent(out):: fjac(n,n)
  real(dp):: tmui(nind)
  psi=x(1:(n-1))
  tau=exp(x(n))
  call update_mu_dmu
  tau=log(tau)
  tmui=1.0_dp+exp(tau)*mui
  fvec(1:(n-1))=sum(Scr,2)
  fvec(n)=sum((exp(tau)*((ni-mui)**2.0_dp - mui*tmui)/(tmui*tmui))*pr_wt)
  fjac(1:(n-1),1:(n-1))=pDVD
  fjac(n,:)=0.0_dp
  fjac(:,n)=0.0_dp
  fjac(n,n)=sum(pr_wt*(((mui/tmui)*exp(tau))**2.0_dp))
end subroutine stepMT

subroutine update_mu_dmu
  use mytype
  use dm_util, only: diagadd
  use matrix, only: operator(.mm.), operator(.tmm.)
  implicit none
  integer:: i, j
  real(dp):: tmpint(4*npan,ntrt), Bhold(4*npan,ntBc), smpsi(nBc,ntrt), nspar(ntrt), &
       & tmui(nind), tpsi(nBc,ntrt), WW(npan,npan), mmu(npan,ntrt), Dmut(npan,ntBc), &
       & inmuij(npan,nind)
  tpsi=reshape(psi,(/nBc,ntrt/))
  tmpint(:,1)=w_grid*exp(Bmat.mm.tpsi(:,1))  
  smpsi(:,1)=(PenM.mm.tpsi(:,1))*spar(1)
  do i=1,nBc
     Bhold(:,i)=Bmat(:,i)*tmpint(:,1)
  end do
  if (ntrt > 1) then
     do i=2,ntrt
        tmpint(:,i)=tmpint(:,1)*exp(Bmat.mm.tpsi(:,i))
        smpsi(:,i)=(PenM.mm.tpsi(:,i))*spar(i)
        do j=((i-1)*nBc+1),(i*nBc)
           Bhold(:,j)=Bmat(:,(j-(i-1)*nBc))*tmpint(:,i)
        end do
     end do
  end if
  do i=1,npan
     mmu(i,:)=sum(tmpint(((i-1)*4+1):(i*4),:),1)
     Dmu(i,:)=sum(Bhold(((i-1)*4+1):(i*4),:),1)
  end do
  do i=1,nind
     muij(:,i)=mmu(:,Itrt(i))
  end do
  where (nijNA)
     muij=0.0_dp
     inmuij=0.0_dp
  elsewhere
     inmuij=1.0_dp/muij
  end where
  DVD=0.0_dp
  mui=sum(muij,1)
  tmui=1.0_dp+tau*mui
  nspar=real(Xvc,dp)*spar
  do i=1,nind
     Dmut=0.0_dp
     WW=-tau/tmui(i)
     call diagadd(WW,inmuij(:,i))
     Dmut(:,1:nBc)=Dmu(:,((Itrt(i)-1)*nBc+1):(Itrt(i)*nBc))
     Dmut(:,((Itrt(i)-1)*nBc+1):(Itrt(i)*nBc))=Dmu(:,((Itrt(i)-1)*nBc+1):(Itrt(i)*nBc))
     Dmut(pack(sqpan,nijNA(:,i)),:)=0.0_dp
     Scr(:,i)=((WW.mm.(nij(:,i)-muij(:,i))).mm.Dmut)*pr_wt(i)
     Scr(((Itrt(i)-1)*nBc+1):(Itrt(i)*nBc),i)=Scr(((Itrt(i)-1)*nBc+1):(Itrt(i)*nBc),i)&
          & -smpsi(:,Itrt(i))
     DVD=DVD+((Dmut.tmm.WW).mm.Dmut)*pr_wt(i)
  end do
  pDVD=DVD
  do i=1,ntrt
     pDVD(((i-1)*nBc+1):(i*nBc),((i-1)*nBc+1):(i*nBc))=pDVD(((i-1)*nBc+1):(i*nBc),&
          & ((i-1)*nBc+1):(i*nBc))+(nspar(i)*PenM)
  end do
end subroutine update_mu_dmu

function TauEst(x)
  use mytype
  use dm_util, only: diagadd
  implicit none
  real(dp), intent(in):: x
  real(dp):: TauEst
  real(dp):: nivar(nind), tmui(nind)
  tau=x
  tmui=1.0_dp+tau*mui
  nivar=mui*tmui
  TauEst=sum(pr_wt*(((ni-mui)**2)/nivar + log(nivar)))
end function TauEst

subroutine tau_update
  use mytype
  use dm_util, only: fmin
  implicit none
  real(dp):: fval
  mui=sum(muij,1)
  call fmin(0.0001_dp,3.0_dp,tau,fval,TauEst,1.0e-2_dp) 
end subroutine tau_update

subroutine s_step(task)
  implicit none
  integer, intent(in):: task
  integer:: i
  do i=1,ngrp
     pr_wt => pr_wt_t(:,i)
     muij => mu_t((((i-1)*npan)+1):(i*npan),1:nind)
     Dmu => Dmu_t((((i-1)*npan)+1):(i*npan),1:ntBc)
     psi => coef_t(1:ntBc,i)
     tau => coef_t(ntBc+1,i)
     spar => coef_t((ntBc+2):(ntBc+ntrt+1),i)
     if (task==0) then
        call init_param
        call param_update(nmaxit)
     elseif (task==1) then
        call param_update(pmaxit)
     else
        call fparam_update(pmaxit)
     end if
     if (err /= 0) return
  end do
end subroutine s_step

subroutine e_step(task)
  use mytype
  use dm_util, only: poisson
  implicit none
  integer, intent(in):: task
  integer:: i, j, k
  real(dp):: dd(npan), vv(nind), gmm
  do i=1,ngrp
     muij => mu_t((((i-1)*npan)+1):(i*npan),1:nind)
     tau => coef_t(ntBc+1,i)
     mui=sum(muij,1)
     if (task==2) then 
        gmm=exp(tau)
        vv=(1.0_dp+gmm*ni)/(1.0_dp+gmm*mui)
     else
        vv=(1.0_dp+tau*ni)/(1.0_dp+tau*mui)
     end if
     do j=1,nind     
        do k=1,npan
           if (nijNA(k,j)) then
              dd(k)=0.0_dp
           else
              dd(k)=poisson(nij(k,j),vv(j)*muij(k,j))
           end if
        end do
        ddlike(j,i)=sum(dd)
        dplike(j,i)=log(probs(i))+ddlike(j,i)
     end do   
  end do
  rscl=maxval(dplike(:,1:ngrp),2)
  do i=1,ngrp
     ddlike(:,i)=exp(ddlike(:,i)-rscl)
     dplike(:,i)=exp(dplike(:,i)-rscl)
  end do
  dlike=sum(dplike(:,1:ngrp),2)
  do i=1,ngrp
     pr_wt_t(:,i)=dplike(:,i)/dlike
  end do
  probs(1:ngrp)=sum(pr_wt_t(:,1:ngrp),1)/real(nind,dp)
end subroutine e_step

subroutine es_update
  use mytype
  use matrix, only: operator(.mm.)
  implicit none
  real(dp):: loglikeo, tpsi(nBc)
  integer:: i, j, k, task
  task=0
  call s_step(task)
  if (err /= 0) return
  call e_step(task)
  loglike=sum(log(dlike)+rscl-gmln)
  do i=1,ngrp
     psi => coef_t(1:ntBc,i)
     spar => coef_t((ntBc+2):(ntBc+ntrt+1),i)
     do j=1,ntrt
        tpsi=psi(((j-1)*nBc+1):(j*nBc))
        loglike=loglike-real(Xvc(j),dp)*spar(j)*sum((PenM.mm.tpsi)*tpsi)/2.0_dp
     end do
  end do
  loglikeo=loglike
  write(*,form1) 'Initial solution   :', loglike
  task=1
  do i=1,nmaxit
     if (i==5) then
        coef_t(ntBc+1,1:ngrp)=log(coef_t(ntBc+1,1:ngrp))
        task=2
     end if
     call s_step(task)
     if (err /= 0) return
     call e_step(task)
     loglike=sum(log(dlike)+rscl-gmln)
     do j=1,ngrp
        psi => coef_t(1:ntBc,j)
        spar => coef_t((ntBc+2):(ntBc+ntrt+1),j)
        do k=1,ntrt
           tpsi=psi(((k-1)*nBc+1):(k*nBc))
           loglike=loglike-real(Xvc(k),dp)*spar(k)*sum((PenM.mm.tpsi)*tpsi)/2.0_dp
        end do
     end do
     if (abs(loglike-loglikeo)<0.01_dp .and. i>5) then
        write(*,form1) 'EM-updated solution:', loglike
        exit
     end if
     loglikeo=loglike
  end do
  write(*,*)
end subroutine es_update

subroutine finalize
  use mytype
  use dm_util, only: newton
  use matrix, only: operator(.mm.)
  implicit none
  integer:: i, j
  real(dp):: tpsi(nBc), coeff(nBgrp), grd(nBgrp), jaco(nBgrp,nBgrp)
  coeff=pack(coef_t(1:(ntBc+1),1:ngrp),.true.)
  if (minval(probs(1:ngrp)) < 1.0_dp/real(nind,dp)) then
     write(*,*) 'Minimal component prob -> 0!'
     err=999
     return
  end if
  call newton(nBgrp,coeff,grd,jaco,stepF,nmaxit,err)
  if (err /= 0) return
  coef_t(ntBc+1,1:ngrp)=exp(coef_t(ntBc+1,1:ngrp))
  loglike=sum(log(dlike)+rscl-gmln)
  do i=1,ngrp
     psi => coef_t(1:ntBc,i)
     spar => coef_t((ntBc+2):(ntBc+ntrt+1),i)
     do j=1,ntrt
        tpsi=psi(((j-1)*nBc+1):(j*nBc))
        loglike=loglike-real(Xvc(j),dp)*spar(j)*sum((PenM.mm.tpsi)*tpsi)/2.0_dp
     end do
  end do
end subroutine finalize

subroutine JackVar(VarCov, jcoef)
  use mytype
  use dm_util, only: newton
  use matrix, only: operator(.mmt.)
  implicit none
  real(dp), intent(out):: VarCov(nBngr,nBngr), jcoef(nBngr,nind)
  real(dp):: coeff(nBngr), grd(nBgrp), jaco(nBgrp,nBgrp), o_Itrt(nind), &
       & o_nij(npan,nind), o_ni(nind), j_coeff(nBngr,nind)
  logical:: o_nijNA(npan,nind)
  integer:: i, indi, j, jind(nind)
  o_nijNA=nijNA
  o_ni=ni
  o_nij=nij
  o_Itrt=Itrt
  jind=(/(i,i=1,nind)/)
  nind=nind-1
  coef_t(ntBc+1,1:ngrp)=log(coef_t(ntBc+1,1:ngrp))
  coeff(1:nBgrp)=pack(coef_t(1:(ntBc+1),1:ngrp),.true.)
  if (ngrp > 1) then
     coeff((nBgrp+1):nBngr)=probs(1:(ngrp-1))
  end if
  do i=1,(nind+1)
     write(*,'(28X,I3)') i
     jind=(/jind(2:(nind+1)),jind(1)/)
     nijNA=o_nijNA(:,jind)
     ni=o_ni(jind)
     nij=o_nij(:,jind)
     Itrt=o_Itrt(jind)
     if (ntrt > 1) then
        do j=1,ntrt
           Xvc(j)=count(Itrt(1:nind)==j)
        end do
     end if
     j_coeff(:,i)=coeff
     call newton(nBgrp,j_coeff(1:nBgrp,i),grd,jaco,stepF,nmaxit,err)
     if (err /= 0) return
     if (ngrp > 1) then
        j_coeff((nBgrp+1):nBngr,i)=probs(1:(ngrp-1))
     end if
  end do
  do i=1,ngrp
     indi=i*(ntBc+1)
     j_coeff(indi,:)=exp(j_coeff(indi,:))
     coeff(indi)=exp(coeff(indi))
  end do
  jcoef=j_coeff
  do i=1,(nind+1)
     j_coeff(:,i)=j_coeff(:,i)-coeff
  end do
  VarCov=((nXr-eff_df)/nXr)*(j_coeff.mmt.j_coeff)
end subroutine JackVar

function OnstVar() result(VarCov)
  use mytype
  use matrix, only: operator(.mmt.), fsymsolve
  implicit none
  real(dp):: VarCov(nBngr,nBngr), o_Itrt(nind), o_nij(npan,nind), o_ni(nind), &
       & o_pr_wt_t(nind,ngrp), o_mu_t(npgrp,nind), o_Scr_t(nBngr,nind), &
       & VC(nBngr,nBngr), Sc(nBngr,nind), o_ddlike(nind,ngrp), o_dlike(nind)
  logical:: o_nijNA(npan,nind)
  integer:: i, j, jind(nind)
  o_nijNA=nijNA
  o_ni=ni
  o_nij=nij
  o_Itrt=Itrt
  o_pr_wt_t=pr_wt_t(:,1:ngrp)
  o_Scr_t=Scr_t(1:nBngr,:)
  o_mu_t=mu_t(1:npgrp,1:nind)
  o_ddlike=ddlike(:,1:ngrp)
  o_dlike=dlike
  jind=(/(i,i=1,nind)/)
  do i=1,nind
     jind=(/jind(2:nind),jind(1)/)
     nijNA=o_nijNA(:,jind)
     ni=o_ni(jind)
     nij=o_nij(:,jind)
     Itrt=o_Itrt(jind)
     if (ntrt > 1) then
        do j=1,ntrt
           Xvc(j)=count(Itrt(1:(nind-1))==j)
        end do
     end if
     pr_wt_t(:,1:ngrp)=o_pr_wt_t(jind,:)
     mu_t(1:npgrp,1:nind)=o_mu_t(:,jind)
     dlike=o_dlike(jind)
     ddlike(:,1:ngrp)=o_ddlike(jind,1:ngrp)
     VC=VarC()
     call fsymsolve(VC,o_Scr_t(:,i),Sc(:,i),err)
     if (err /=0 ) return
  end do
  VarCov=((nXr-eff_df)/nXr)*(Sc.mmt.Sc)
  nijNA=o_nijNA
  ni=o_ni
  nij=o_nij
  Itrt=o_Itrt
end function OnstVar

function VarC() result(VarCov)
  use mytype
  use dm_util, only: diagadd, outerprod
  use matrix, only: operator(.mm.), operator(.tmm.), operator(.mmt.), fsymsolve
  implicit none
  real(dp):: nspar(ntrt), VarCov(nBngr,nBngr), tmui(nind-1), WW(npan,npan), &
       & dDVD(ntBc+1,ntBc+1), UU(ntBc,nind-1), sUU(ntBc,nind-1), dH(nind-1), Dmut(npan,ntBc), &
       & inmuij(npan,nind-1)
  integer:: i, j, nBco, tnind
  nBco=ntBc+1
  tnind=nind-1
  VarCov=0.0_dp
  do i=1,ngrp
     pr_wt => pr_wt_t(1:tnind,i)
     muij => mu_t((((i-1)*npan)+1):(i*npan),1:tnind)
     Dmu => Dmu_t((((i-1)*npan)+1):(i*npan),1:ntBc)
     Scr_p => Scr_t((((i-1)*nBco)+1):(i*nBco),1:tnind)
     tau => coef_t(ntBc+1,i)
     spar => coef_t((ntBc+2):(ntBc+ntrt+1),i)
     nspar=real(Xvc,dp)*spar
     DVD=0.0_dp
     dDVD=0.0_dp
     where (nijNA(:,1:tnind))
        muij=0.0_dp
        inmuij=0.0_dp
     elsewhere
        inmuij=1.0_dp/muij
     end where
     mui(1:tnind)=sum(muij,1)
     tmui=1.0_dp+tau*mui(1:tnind)
     do j=1,tnind
        Dmut=0.0_dp
        WW=-tau/tmui(j)
        call diagadd(WW,inmuij(:,j))
        Dmut(:,1:nBc)=Dmu(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))
        Dmut(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))=Dmu(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))
        Dmut(pack(sqpan,nijNA(:,j)),:)=0.0_dp
        Scr_p(1:ntBc,j)=((WW.mm.(nij(:,j)-muij(:,j))).mm.Dmut)
        DVD=DVD+((Dmut.tmm.WW).mm.Dmut)*pr_wt(j)
        UU(:,j)=sum(Dmut,1)/tmui(j)
     end do
     pDVD=DVD
     do j=1,ntrt
        pDVD(((j-1)*nBc+1):(j*nBc),((j-1)*nBc+1):(j*nBc))=pDVD(((j-1)*nBc+1):(j*nBc),&
             & ((j-1)*nBc+1):(j*nBc))+(nspar(j)*PenM)
     end do
     call fsymsolve(pDVD,UU,sUU,err)
     if (err /=0) return
     dH=sum(UU*sUU,1)
     Scr_p(nBco,1:tnind)=(((ni(1:tnind)-mui(1:tnind))**2.0_dp - mui(1:tnind)*tmui)/(tmui*tmui) + dH)
     do j=1,tnind
        dDVD=dDVD+outerprod(Scr_p(:,j),Scr_p(:,j))*pr_wt(j)
        Scr_p(:,j)=Scr_p(:,j)*pr_wt(j)
     end do
     VarCov((((i-1)*nBco)+1):(i*nBco),(((i-1)*nBco)+1):(i*nBco))=-dDVD
     VarCov((((i-1)*nBco)+1):((i*nBco)-1),(((i-1)*nBco)+1):((i*nBco)-1))=VarCov(((&
          & (i-1)*nBco)+1):((i*nBco)-1),(((i-1)*nBco)+1):((i*nBco)-1))+pDVD
     VarCov((i*nBco),(i*nBco))=VarCov((i*nBco),(i*nBco))+sum(pr_wt*((mui(1:tnind)/tmui)**2.0_dp+sum(dH*dH)))
     ddlike(1:tnind,i)=ddlike(1:tnind,i)/dlike(1:tnind)
  end do
  if (ngrp > 1) then
     do i=1,(ngrp-1)
        ddlike(1:tnind,i)=ddlike(1:tnind,i)-ddlike(1:tnind,ngrp)
        VarCov(((i-1)*nBco+1):(i*nBco),nBgrp+i)=-sum(Scr_t((((i-1)*nBco)+1):(i*nBco),1:tnind),2)/probs(i)
        VarCov(((ngrp-1)*nBco+1):nBgrp,nBgrp+i)=sum(Scr_t((((ngrp-1)*nBco)+1):nBgrp,1:tnind),2)/probs(ngrp)
     end do
     Scr_t((nBgrp+1):(nBngr),1:tnind)=transpose(ddlike(1:tnind,1:(ngrp-1)))
     VarCov=(Scr_t(1:nBngr,1:tnind).mmt.Scr_t(1:nBngr,1:tnind))+VarCov
     VarCov((nBgrp+1):(nBngr),1:nBgrp)=transpose(VarCov(1:nBgrp,(nBgrp+1):&
          &(nBngr)))
  else
     VarCov=(Scr_t(1:nBngr,1:tnind).mmt.Scr_t(1:nBngr,1:tnind))+VarCov
  end if
end function VarC

function SandVar() result(VarCov)
  use mytype
  use dm_util, only: diagadd, outerprod
  use matrix, only: fsymsolve, diag, operator(.mm.), operator(.tmm.), operator(.mmt.), finverse
  implicit none
  real(dp):: smpsi(ntBc,ngrp), nspar(ntrt), VarCov(nBngr,nBngr), Scb(nBngr,nBngr), &
       & tmui(nind), WW(npan,npan), dDVD(ntBc+1,ntBc+1), UU(ntBc,nind), sUU(ntBc,nind), &
       & dH(nind), Dmut(npan,ntBc), tpsi(nBc,ntrt), tDVD(ntBc,ntBc), t_ddlike(nind,ngrp), &
       & inmuij(npan,nind)
  integer:: i, j, nBco
  nBco=ntBc+1
  VarCov=0.0_dp
  eff_df=0.0_dp
  do i=1,ngrp
     pr_wt => pr_wt_t(:,i)
     muij => mu_t((((i-1)*npan)+1):(i*npan),1:nind)
     Dmu => Dmu_t((((i-1)*npan)+1):(i*npan),1:ntBc)
     Scr_p => Scr_t((((i-1)*nBco)+1):(i*nBco),1:nind)
     psi => coef_t(1:ntBc,i)
     tau => coef_t(ntBc+1,i)
     spar => coef_t((ntBc+2):(ntBc+ntrt+1),i)
     tpsi=reshape(psi,(/nBc,ntrt/))
     do j=1,ntrt
        smpsi(((j-1)*nBc+1):(j*nBc),i)=(PenM.mm.tpsi(:,j))*spar(j)
     end do
     nspar=real(Xvc,dp)*spar
     DVD=0.0_dp
     dDVD=0.0_dp
     where (nijNA(:,1:nind))
        muij=0.0_dp
        inmuij=0.0_dp
     elsewhere
        inmuij=1.0_dp/muij
     end where
     mui=sum(muij,1)
     tmui=1.0_dp+tau*mui
     do j=1,nind
        Dmut=0.0_dp
        WW=-tau/tmui(j)
        call diagadd(WW,inmuij(:,j))
        Dmut(:,1:nBc)=Dmu(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))
        Dmut(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))=Dmu(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))
        Dmut(pack(sqpan,nijNA(:,j)),:)=0.0_dp
        Scr_p(1:ntBc,j)=((WW.mm.(nij(:,j)-muij(:,j))).mm.Dmut)
        DVD=DVD+((Dmut.tmm.WW).mm.Dmut)*pr_wt(j)
        UU(:,j)=sum(Dmut,1)/tmui(j)
     end do
     pDVD=DVD
     do j=1,ntrt
        pDVD(((j-1)*nBc+1):(j*nBc),((j-1)*nBc+1):(j*nBc))=pDVD(((j-1)*nBc+1):(j*nBc),&
             & ((j-1)*nBc+1):(j*nBc))+(nspar(j)*PenM)
     end do
     call fsymsolve(pDVD(1:ntBc,1:ntBc),DVD,tDVD,err)
     if (err /= 0) return
     eff_df=eff_df+sum(diag(tDVD))
     call fsymsolve(pDVD,UU,sUU,err)
     if (err /= 0) return
     dH=sum(UU*sUU,1)
     Scr_p(nBco,:)=(((ni-mui)**2.0_dp - mui*tmui)/(tmui*tmui) + dH)
     do j=1,nind
        dDVD=dDVD+outerprod(Scr_p(:,j),Scr_p(:,j))*pr_wt(j)
        Scr_p(:,j)=Scr_p(:,j)*pr_wt(j)
     end do
     VarCov((((i-1)*nBco)+1):(i*nBco),(((i-1)*nBco)+1):(i*nBco))=-dDVD
     VarCov((((i-1)*nBco)+1):((i*nBco)-1),(((i-1)*nBco)+1):((i*nBco)-1))=VarCov(((&
          & (i-1)*nBco)+1):((i*nBco)-1),(((i-1)*nBco)+1):((i*nBco)-1))+pDVD
     VarCov((i*nBco),(i*nBco))=VarCov((i*nBco),(i*nBco))+sum(pr_wt*((mui/tmui)**2.0_dp+sum(dH*dH)))
     t_ddlike(:,i)=ddlike(:,i)/dlike
  end do
  if (ngrp > 1) then
     do i=1,(ngrp-1)
        t_ddlike(:,i)=t_ddlike(:,i)-t_ddlike(:,ngrp)
        VarCov(((i-1)*nBco+1):(i*nBco),nBgrp+i)=-sum(Scr_t((((i-1)*nBco)+1):(i*nBco),:),2)/probs(i)
        VarCov(((ngrp-1)*nBco+1):nBgrp,nBgrp+i)=sum(Scr_t((((ngrp-1)*nBco)+1):nBgrp,:),2)/probs(ngrp)
     end do
     Scr_t((nBgrp+1):(nBngr),:)=transpose(t_ddlike(:,1:(ngrp-1)))
     VarCov=(Scr_t(1:nBngr,:).mmt.Scr_t(1:nBngr,:))+VarCov
     VarCov((nBgrp+1):(nBngr),1:nBgrp)=transpose(VarCov(1:nBgrp,(nBgrp+1):&
          &(nBngr)))
  else
     VarCov=(Scr_t(1:nBngr,:).mmt.Scr_t(1:nBngr,:))+VarCov
  end if
  do i=1,ngrp
     Scr_p => Scr_t((((i-1)*nBco)+1):(i*nBco),1:nind)
     do j=1,nind        
        Scr_p(((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc),j)=Scr_p(((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc),j) &
             & -smpsi(((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc),i)
     end do
  end do
  eff_df=eff_df+2.0_dp*ngrp-1.0_dp
  Scb=Scr_t(1:nBngr,:).mmt.Scr_t(1:nBngr,:)
  VarCov=finverse(VarCov,err)
  if (err/=0) return
  VarCov=VarCov.mm.(Scb.mmt.VarCov)
end function SandVar

subroutine stepF(n,x,fvec,fjac)
  use mytype
  use dm_util, only: diagadd, outerprod, poisson
  use matrix, only: fsymsolve, operator(.mm.), operator(.tmm.)
  implicit none
  integer, intent(in):: n
  real(dp), intent(in):: x(n)
  real(dp), intent(out):: fvec(n)
  real(dp), intent(out):: fjac(n,n)
  real(dp):: tmpint(4*npan,ntrt), Bhold(4*npan,ntBc), smpsi(nBc,ntrt), ddd(npan), &
       & nspar(ntrt), vv(nind), WW(npan,npan), tmui(nind), UU(ntBc,nind), &
       & sUU(ntBc,nind), dH(nind), Dmut(npan,ntBc), tpsi(nBc,ntrt), mmu(npan,ntrt), &
       & inmuij(npan,nind)
  integer:: i, j, k, nBco
  nBco=ntBc+1
  coef_t(1:nBco,1:ngrp)=reshape(x(1:nBgrp),(/nBco,ngrp/))
  do i=1,ngrp
     psi => coef_t(1:ntBc,i)
     tau => coef_t(ntBc+1,i)
     muij => mu_t((((i-1)*npan)+1):(i*npan),1:nind)
     Dmu => Dmu_t((((i-1)*npan)+1):(i*npan),1:ntBc)
     tpsi=reshape(psi,(/nBc,ntrt/))
     tmpint(:,1)=w_grid*exp(Bmat.mm.tpsi(:,1))  
     do j=1,nBc
        Bhold(:,j)=Bmat(:,j)*tmpint(:,1)
     end do
     if (ntrt > 1) then
        do j=2,ntrt
           tmpint(:,j)=tmpint(:,1)*exp(Bmat.mm.tpsi(:,j))
           do k=((j-1)*nBc+1),(j*nBc)
              Bhold(:,k)=Bmat(:,(k-(j-1)*nBc))*tmpint(:,j)
           end do
        end do
     end if
     do j=1,npan
        mmu(j,:)=sum(tmpint(((j-1)*4+1):(j*4),:),1)
        Dmu(j,:)=sum(Bhold(((j-1)*4+1):(j*4),:),1)
     end do
     do j=1,nind
        muij(:,j)=mmu(:,Itrt(j))
     end do
     where (nijNA(:,1:nind))
        muij=0.0_dp
     end where
     mui(1:nind)=sum(muij,1)
     vv=(1.0_dp+exp(tau)*ni(1:nind))/(1.0_dp+exp(tau)*mui(1:nind))
     do j=1,nind     
        do k=1,npan
           if (nijNA(k,j)) then
              ddd(k)=0.0_dp
           else
              ddd(k)=poisson(nij(k,j),vv(j)*muij(k,j))
           end if
        end do
        ddlike(j,i)=sum(ddd)
        dplike(j,i)=log(probs(i))+ddlike(j,i)
     end do
  end do
  rscl(1:nind)=maxval(dplike(1:nind,1:ngrp),2)
  do i=1,ngrp
     ddlike(1:nind,i)=exp(ddlike(1:nind,i)-rscl(1:nind))
     dplike(1:nind,i)=exp(dplike(1:nind,i)-rscl(1:nind)) 
  end do
  dlike(1:nind)=sum(dplike(1:nind,1:ngrp),2)
  fjac=0.0_dp
  do i=1,ngrp
     pr_wt_t(1:nind,i)=dplike(1:nind,i)/dlike(1:nind)
     pr_wt => pr_wt_t(1:nind,i)
     muij => mu_t((((i-1)*npan)+1):(i*npan),1:nind)
     Dmu => Dmu_t((((i-1)*npan)+1):(i*npan),1:ntBc)
     Scr_p => Scr_t((((i-1)*nBco)+1):(i*nBco),1:nind)
     psi => coef_t(1:ntBc,i)
     tau => coef_t(ntBc+1,i)
     spar => coef_t((ntBc+2):(ntBc+ntrt+1),i)
     tpsi=reshape(psi,(/nBc,ntrt/))
     do j=1,ntrt
        smpsi(:,j)=(PenM.mm.tpsi(:,j))*spar(j)
     end do
     nspar=real(Xvc,dp)*spar
     where (nijNA(:,1:nind))
        muij=0.0_dp
        inmuij=0.0_dp
     elsewhere
        inmuij=1.0_dp/muij
     end where
     mui(1:nind)=sum(muij,1)
     tmui=1.0_dp+exp(tau)*mui(1:nind)
     DVD=0.0_dp
     do j=1,nind
        Dmut=0.0_dp
        WW=-exp(tau)/tmui(j)
        call diagadd(WW,inmuij(:,j))
        Dmut(:,1:nBc)=Dmu(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))
        Dmut(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))=Dmu(:,((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc))
        Dmut(pack(sqpan,nijNA(:,j)),:)=0.0_dp
        Scr_p(1:ntBc,j)=((WW.mm.(nij(:,j)-muij(:,j))).mm.Dmut)*pr_wt(j)
        Scr_p(((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc),j)=Scr_p(((Itrt(j)-1)*nBc+1):(Itrt(j)*nBc),j)&
             & -smpsi(:,Itrt(j))
        DVD=DVD+((Dmut.tmm.WW).mm.Dmut)*pr_wt(j)
        UU(:,j)=sum(Dmut,1)/tmui(j)
     end do
     pDVD=DVD
     do j=1,ntrt
        pDVD(((j-1)*nBc+1):(j*nBc),((j-1)*nBc+1):(j*nBc))=pDVD(((j-1)*nBc+1):(j*nBc),&
             & ((j-1)*nBc+1):(j*nBc))+(nspar(j)*PenM)
     end do
     call fsymsolve(pDVD,UU,sUU,err)
     if (err /=0) return
     dH=sum(UU*sUU,1)
     Scr_p(nBco,1:nind)=(exp(tau)*((ni(1:nind)-mui(1:nind))**2.0_dp - mui(1:nind)*tmui)/(tmui*tmui) + dH)*pr_wt
     fjac((((i-1)*nBco)+1):((i*nBco)-1),(((i-1)*nBco)+1):((i*nBco)-1))=pDVD
     fjac((i*nBco),(i*nBco))=sum(pr_wt*(((mui(1:nind)/tmui)*exp(tau))**2.0_dp))
  end do
  fvec=sum(Scr_t(1:n,1:nind),2)
  probs(1:ngrp)=sum(pr_wt_t(1:nind,1:ngrp),1)/real(nind,dp)
  if (minval(probs(1:ngrp)) < 2.0_dp/real(nind,dp)) then
     write(*,*) 'Minimal component prob -> 0!'
     err=999
  end if
end subroutine stepF

subroutine dmNHPP(panel,datij,Xm,coefs,pwt,VrCov,tcoef,llk,edf,npn,nsub,nknt,ntr,ngr,fin,vind,info)
  use mytype
  implicit none
  integer, intent(in):: npn, nsub, nknt, ntr, ngr, fin, vind
  real(dp), intent(in):: panel(npn), datij(npn,nsub)
  integer, intent(in):: Xm(ntr)
  real(dp), intent(out):: coefs(ntr*(nknt+5)+1,ngr), llk, edf, &
       & VrCov(3*(ngr*(ntr*(nknt+4)+1)+ngr-1),ngr*(ntr*(nknt+4)+1)+ngr-1), &
       & tcoef((ngr*(ntr*(nknt+4)+1)+ngr-1),nsub)
  real(dp), intent(inout):: pwt(nsub,ngr)
  integer, intent(out):: info
  real(dp):: VC(ngr*(ntr*(nknt+4)+1)+ngr-1,ngr*(ntr*(nknt+4)+1)+ngr-1)
  write(*,'(A20)') '|---Initializing---:'
  call initialize(datij,Xm,panel,pwt,npn,nsub,nknt,ntr,ngr,fin)
  if (err /=0) then
     info=err
     return
  end if
  write(*,'(A33)') '                         [DONE]'
  call es_update
  if (err /= 0) then
     call unload
     info=err
     return
  end if
  if (fin == 1) then
     write(*,'(A20)') '|-----Refining-----:'
     call finalize
     write(*,'(A33)') '                         [DONE]'
     if (err == 0) then
        coefs=coef_t(:,1:ngrp)
        pwt=pr_wt_t(:,1:ngrp)
        llk=loglike
        if (vind == 1) then
           VrCov(1:nBngr,:)=SandVar()
           edf=eff_df
           if (err /=0) then
              call unload
              info=err        
              return
           end if
           write(*,'(A20)') '|----Jack-Knife----:'
           VrCov((2*nBngr+1):3*nBngr,:)=OnstVar()
           if (err /=0) then
              call unload
              info=err
              return
           end if
           call JackVar(VC,tcoef)
           VrCov((nBngr+1):2*nBngr,:)=VC
           if (err /= 0) then
              call unload
              info=err
              return
           end if
           write(*,'(A33)') '                      [DONE]'
        else
           VrCov=0.0_dp
        end if
     else
        info=err
     end if
  else
     if (err == 0) then
        coefs=coef_t(:,1:ngrp)
        coefs(ntBc+1,1:ngrp)=exp(coef_t(ntBc+1,1:ngrp))
        pwt=pr_wt_t(:,1:ngrp)
        llk=loglike
     else
        info=err
     end if
  end if
  write(*,form1) 'Final solution     :', loglike
  call unload
end subroutine dmNHPP

end module dm_NHPP_mod

subroutine R_dmNHPP(panel,datij,Xm,coefs,pwt,VrCov,tcoef,llk,edf,npn,nsub,nknt,ntr,ngr,fin,vind,pini,info)
  use mytype
  use stp_dm_NHPP_mod
  use dm_NHPP_mod
  implicit none
  integer, intent(in):: npn, nsub, nknt, ntr, ngr, fin, vind, pini
  real(dp), intent(in):: panel(npn), datij(npn,nsub)
  integer, intent(in):: Xm(ntr)
  real(dp), intent(out):: coefs(ntr*(nknt+5)+1,ngr), llk, edf, &
       & VrCov(3*(ngr*(ntr*(nknt+4)+1)+ngr-1),ngr*(ntr*(nknt+4)+1)+ngr-1), &
       & tcoef((ngr*(ntr*(nknt+4)+1)+ngr-1),nsub)
  real(dp), intent(inout):: pwt(nsub,ngr)
  integer, intent(out):: info
  if (pini==1) then
     call stpdmNHPP(datij,Xm,panel,pwt,nsub,npn,ntr,ngr,info)
     if (info /= 0) then
        return
     end if
  end if
  call dmNHPP(panel,datij,Xm,coefs,pwt,VrCov,tcoef,llk,edf,npn,nsub,nknt,ntr,ngr,fin,vind,info)
end subroutine R_dmNHPP
